#!python

from typing import Dict, List, Union, Callable, Any
from magLabUtilities.optimizerutilities.parameterSpaces import ParameterSpace

class FitKernel:
    def __init__(
                    self,
                    refFP:str,
                    parameterDefs:Dict[str, Dict[str, Any]], 
                    optimizerConstructorMethod:Callable[...,Callable],
                    optimizerConfig:Dict[str, Dict[str, Any]],
                    tuneHistoryFP:str
                ):

        self.parameterSpace = ParameterSpace(parameterDefs)


#!python3

#\TODO implement message pipe as a static class to allow calls without instantiation

class MP:
    def __init__(self, initMessage, messageTypes):
        # Stores the names (as strings) of each active message type
        self.messageTypesDict = messageTypes

        # Prints a custom message at initialization
        print('\n' + initMessage)

        if self.messageTypesDict.keys() == []:
            raise Exception('Message Pipe: No message types specified in initializer.')
        elif 'All' in self.messageTypesDict.keys() and 'None' in self.messageTypesDict.keys():
            raise Exception('Message Pipe: Cannot have both "All" and "None" messages.')
        elif 'All' in self.messageTypesDict.keys():
            if self.messageTypesDict['All'] == 1:
                for messageType in self.messageTypesDict:
                    self.messageTypesDict[messageType] = 1
        elif 'None' in self.messageTypesDict.keys():
            if self.messageTypesDict['None'] == 1:
                for messageType in self.messageTypesDict:
                    self.messageTypesDict[messageType] = 0

        # Lists all configured message types and whether they are enabled
        print ('\n    Message Types:')
        for messageType in self.messageTypesDict:
            if self.messageTypesDict[messageType] == 1:
                print('        %s: Enabled ' % '{:13}'.format(messageType))
            else:
                print('        %s: Disabled ' % '{:13}'.format(messageType))
        print('')

    # Receives a message and decides whether to display it on the terminal
    def out(self, messageTypes, message):
        if self.messageTypesDict == {}:
            raise Exception('Message Pipe: Pipe is not initialized.')
        else:
            for messageType in messageTypes:
                if messageType in self.messageTypesDict.keys():
                    if self.messageTypesDict[messageType] == 1:
                        print('%s>> %s' % ('{:13}'.format(messageType), message))
                else:
                    raise Exception('Message Pipe: Message type "%s" not recognized.' % messageType)


#!python3

import numpy as np
from pathos.threading import ThreadPool
from typing import Callable, List, Dict, Union
from magLabUtilities.signalutilities.signals import SignalThread, Signal, SignalBundle
from magLabUtilities.optimizerutilities.parameterSpaces import TestGrid, GridNode

class GradientDescent:
    def __init__(self, parameterList:List[Dict[str,Union[str, float, List[int]]]], costFunction:Callable[[GridNode], None], gradientStepFunction:Callable[[GridNode],None]):
        self.parameterList = parameterList
        self.costFunction = costFunction
        self.gradientStepFunction = gradientStepFunction

        self.testGrid = TestGrid(parameterList)

    def tune(self, numIterations:int=np.infty, maxThreads=None):
        nodeIndices = [0] * len(self.parameterList)
        lowestLossNode = None
        iterationNum = 0
        oldLowestLoss = np.infty

        while iterationNum < numIterations:
            iterationNum += 1
            testGridNodeList = self.testGrid.getTestGridNodes(nodeIndices)

            costFunctionQueue = []
            for testGridNode in testGridNodeList:
                if testGridNode.loss == None:
                    costFunctionQueue.append(testGridNode)

            with ThreadPool(nodes=maxThreads) as workerPool:
                testGridNodeList = workerPool.map(self.costFunction, costFunctionQueue)

            loss = []
            for testGridNode in testGridNodeList:
                loss.append(testGridNode.loss)

            lowestLossNode = testGridNodeList[np.argmin(np.array(loss))]
            self.gradientStepFunction(lowestLossNode)
            if oldLowestLoss <= lowestLossNode.loss:
                break
            else:
                oldLowestLoss = lowestLossNode.loss
                nodeIndices = lowestLossNode.indexList



#!python3

import numpy as np
from typing import List
from magLabUtilities.signalutilities.signals import SignalThread, Signal
# from internal.terminalInterface import printMsg

def differentialDerivative(fNum, fDenom, windowRadius, discontinuousPoints, terminalMode='verbose'):
    dF = []
    for index in range(len(fNum)):
        windowIndices = [None] * 2
        for radius in range(windowRadius + 1):
            if index + radius in discontinuousPoints or index - radius in discontinuousPoints:
                if radius == 0:
                    if index == 0:
                        windowIndices[0] = index
                        windowIndices[1] = index + 1
                    else:
                        windowIndices[0] = index - 1
                        windowIndices[1] = index
                else:
                    windowIndices[0] = index - radius
                    windowIndices[1] = index + radius
        if windowIndices == [None] * 2:
            windowIndices[0] = index - windowRadius
            windowIndices[1] = index + windowRadius
        dF.append((fNum[windowIndices[1]] - fNum[windowIndices[0]]) / (fDenom[windowIndices[1]] - fDenom[windowIndices[0]]))
        # printMsg('Index: %d    F: %f    Left Index: %d    F(left): %f    Right Index: %d    F(right):    %f    dF: %f' % (index, fNum[index], windowIndices[0], fNum[windowIndices[0]], windowIndices[1], fNum[windowIndices[1]], dF[index]), terminalMode)
    return dF

def finiteDiffDerivative(fNum:np.ndarray, fDenom:np.ndarray, windowRadius:np.int, discontinuousPoints:List[np.int], differenceMode:str) -> np.ndarray:
    fNum = np.array(fNum, dtype=np.float32)
    fDenom = np.array(fDenom, dtype=np.float32)
    discontinuousPoints = discontinuousPoints + [0, fNum.size-1]
    dF = []
    for index in range(len(fNum)):
        windowIndices = [None] * 2
        for radius in range(windowRadius + 1):
            if index + radius in discontinuousPoints or index - radius in discontinuousPoints:
                if radius == 0:
                    if index == 0:
                        windowIndices[0] = index
                        windowIndices[1] = index + 1
                    else:
                        windowIndices[0] = index - 1
                        windowIndices[1] = index
                else:
                    windowIndices[0] = index - radius
                    windowIndices[1] = index + radius
        if windowIndices == [None] * 2:
            windowIndices[0] = index - windowRadius
            windowIndices[1] = index + windowRadius
        if differenceMode == 'centralDifference':
            dF.append((fNum[windowIndices[1]] - fNum[windowIndices[0]]) / (fDenom[windowIndices[1]] - fDenom[windowIndices[0]]))
        if differenceMode == 'backwardsDifference':
            if index == 0:
                dF.append((fNum[1] - fNum[windowIndices[0]]) / (fDenom[1] - fDenom[windowIndices[0]]))
            else:
                dF.append((fNum[index] - fNum[windowIndices[0]]) / (fDenom[index] - fDenom[windowIndices[0]]))
        # print ('Index: %d    F: %e    Left Index: %d    F(left): %e    Right Index: %d    F(right):    %e    dF: %e' % (index, fNum[index], windowIndices[0], fNum[windowIndices[0]], windowIndices[1], fNum[windowIndices[1]], dF[index]))
    return np.array(dF, dtype=np.float32)

def integralIndexQuadrature(independentThread:np.ndarray, dependentThread:np.ndarray, c:np.float64=0.0) -> SignalThread:
    dT = np.ediff1d(dependentThread, to_begin=dependentThread[0])
    return SignalThread(np.cumsum(independentThread.data * dT) + c)

def integralTrapQuadrature(independentThread:SignalThread, dependentThread:SignalThread, c:np.float64=0.0) -> SignalThread:
    dT = np.ediff1d(dependentThread.data)
    dX = np.ediff1d(independentThread.data)

    outputArray = np.array(np.zeros_like(independentThread.data), dtype=np.float64)
    outputArray[1:] = dT * independentThread.data[:-1] + 0.5 * dT * dX

    return SignalThread(np.cumsum(outputArray) + c)

# integral(data, a, b, quadratureMode=['dataPoints', 'quadrature'], params)

# derivative(xData, yData, differentiationMode=['legendre', 'slopeControlledWindow'], singularityMode=['legendre', 'adaptiveWindow'])



#!python3

class Error(Exception):
    pass

class SignalTypeError(TypeError):
    pass

class SignalValueError(ValueError):
    pass

class FileIOValueError(ValueError):
    pass

class CostValueError(ValueError):
    pass

class UITypeError(TypeError):
    pass

class UIValueError(ValueError):
    pass



#!python3

# \name
#   Function Generator module
#
# \description
#   Generates combinations of canonical waveforms
#
# \notes
#   parabola(self, xMaxStep, xMinStep, x0, x1, power, numPoints)
#       where: 
#           xMaxStep  - x-point with the largest step size (e.g. at saturation)
#           xMinStep  - x-point with the smallest step size (e.g. at coercive field)
#           power     - exponent on the parabola (fractional powers are acceptable)
#           x0        - beginning of the sweep
#           x1        - end of the sweep
#           numPoints - is the number of points in the sweep (inclusive)
#
#   sinusoid(self, amplitude, thetaStep, angularPeriod, numCycles, phaseShift=0.0)
#       where:
#           amplitude       - Amplitude of the sine wave
#           thetaStep       - step size of theta
#           angularPeriod   - angular period of sine wave (2pi is one cylce)
#           numCycles       - number of cycles the sine wave goes through
#           (phaseShift)    - shifts the sine wave (default is 0.0)
#
# \todo
#
# \revised
#   Mark Travers 4/22/2020      - Original construction
#   Stephen Gedney 4/27/2020    - Added Linear function
#   Mark Travers 6/16/2020      - Implemented function generators with Signals. Vectorized Parabola

import numpy as np
from typing import Tuple, Callable
from magLabUtilities.signalutilities.signals import SignalThread, Signal
from magLabUtilities.exceptions.exceptions import SignalValueError

class Line:
    def __init__(self, x0:np.float64, x1:np.float64, t0:np.float64, t1:np.float64, enforceTBounds=True):
        self.x0 = x0
        self.x1 = x1
        self.t0 = t0
        self.t1 = t1
        self.enforceTBounds = enforceTBounds

        self.slope = (self.x1-self.x0) / (self.t1-self.t0)
        self.xIntercept = self.x0 - self.slope * self.t0

    def evaluate(self, tThread:SignalThread) -> Signal:
        if self.enforceTBounds:
            if not (np.all(tThread.data >= self.t0) and np.all(tThread.data <= self.t1)):
                raise SignalValueError('Cannot evaluate function outside bounds.')
        if not tThread.isIncreasing:
            raise SignalValueError('tThread must be increasing.')

        return Signal.fromThreadPair(SignalThread(tThread.data * self.slope + self.xIntercept), tThread)

class Parabola:
    def __init__(self, tMaxDx:np.float64, tVertex:np.float64, xMaxDx:np.float64, xVertex:np.float64, power:np.float64, enforceTBounds=True):
        self.tMaxDx = np.float64(tMaxDx)
        self.tVertex = np.float64(tVertex)
        self.xMaxDx = np.float64(xMaxDx)
        self.xVertex = np.float64(xVertex)
        self.power = np.float64(power)
        self.enforceTBounds = enforceTBounds

    def evaluate(self, tThread:SignalThread) -> Signal:
        if not tThread.isIncreasing:
            raise SignalValueError('tThread must be increasing.')

        if self.tVertex > self.tMaxDx:
            inRegion = np.where(tThread.data >= self.tVertex)
            outRegion = np.where(tThread.data < self.tVertex)

            inThread = -(self.xMaxDx-self.xVertex) * np.power((-tThread.data[inRegion]+self.tVertex)/(self.tMaxDx-self.tVertex), self.power) + self.tVertex
            outThread = (self.xMaxDx-self.xVertex) * np.power((tThread.data[outRegion]-self.tVertex)/(self.tMaxDx-self.tVertex), self.power) + self.tVertex

            xThread = np.hstack((outThread, inThread))

        elif self.tVertex < self.tMaxDx:
            inRegion = np.where(tThread.data <= self.tVertex)
            outRegion = np.where(tThread.data > self.tVertex)

            inThread = -(self.xMaxDx-self.xVertex) * np.power((tThread.data[inRegion]-self.tVertex)/(self.tMaxDx-self.tVertex), self.power) + self.tVertex
            outThread = (self.xMaxDx-self.xVertex) * np.power((-tThread.data[outRegion]+self.tVertex)/(self.tMaxDx-self.tVertex), self.power) + self.tVertex

            xThread = np.hstack((inThread, outThread))

        return Signal.fromThreadPair(SignalThread(xThread), tThread)

class Sinusoid:
    def __init__(self, amplitude:np.float64, angularPeriod:np.float64, phaseShift:np.float64):
        self.amplitude = amplitude
        self.angularPeriod = angularPeriod
        self.phaseShift = phaseShift

    def evaluate(self, tThread:SignalThread) -> Signal:
        return SignalThread(self.amplitude * np.sin(tThread.data - self.phaseShift))

class SeriesApproximation:
    pass



#!python3

import json
import numpy as np
from typing import Dict, List, Tuple, Union

# class ParameterSpace:
#     def __init__(self, parameterDefs:Dict[str, Dict[str, Union[float, List[float]]]]):
#         self.parameterDefs = parameterDefs

class TestGrid:
    def __init__(self, parameterList:List[Dict[str,Union[str, float, List[int]]]]):
        self.parameterList = parameterList

        self.testGridIndices = []
        self._globalNodeDict = {}

        self.buildLocalGridIndices([], 0)

    def getTestGridNodes(self, currentNodeIndices):
        testGridNodeList = []
        for testGridNode in self.testGridIndices:

            nodeIndexList = [testGridNode[i] + currentNodeIndices[i] for i in range(len(currentNodeIndices))]
            nodeIndexKey = json.dumps(nodeIndexList)
            if nodeIndexKey not in list(self._globalNodeDict.keys()):
                self._globalNodeDict[nodeIndexKey] = GridNode(nodeIndexList, self.parameterList)
            testGridNodeList.append(self._globalNodeDict[nodeIndexKey])
        return testGridNodeList

    def buildLocalGridIndices(self, indexList:List[int], parameterIndex:int) -> None:
        for localIndex in self.parameterList[parameterIndex]['testGridLocalIndices']:
            indexList.append(localIndex)
            if parameterIndex < len(self.parameterList) - 1:
                self.buildLocalGridIndices(indexList, parameterIndex + 1)
            else:
                self.testGridIndices.append(indexList)
            indexList = indexList[:parameterIndex]

    @property
    def globalNodeDict(self):
        return self._globalNodeDict

    @property
    def testGridNodeNum(self):
        return len(self.testGridIndices)

class GridNode:
    def __init__(self, indexList:List[np.float64], parameterList):
        self._indexList = indexList
        self._coordList = []
        for index, parameter in enumerate(parameterList):
            self._coordList.append(parameter['initialValue'] + parameter['stepSize'] * float(self._indexList[index]))
        self._loss = None
        self._refPlotBundle = None
        self._testPlotBundle = None
        self.data = {}

        # self.mp.out(['GridNode'], 'Created GridNode with %s and %s' % (self._indexList, self._coordList))

    @property
    def indexList(self):
        return(self._indexList)

    @property
    def coordList(self):
        return(self._coordList)

    @property
    def loss(self):
        return self._loss

    @loss.setter
    def loss(self, loss):
        self._loss = loss



#!python3

import numpy as np
from magLabUtilities.signalutilities.signals import Signal
from magLabUtilities.exceptions.exceptions import CostValueError

def rmsNdNorm(refMatrix:np.ndarray, testMatrix:np.ndarray, tWeightMatrix:np.ndarray=None, normalizeDataByDimRange=True) -> np.float64:
    if (refMatrix.shape != testMatrix.shape) or (not np.all(refMatrix[0,:] == testMatrix[0,:])):
        raise CostValueError('refMatrix and testMatrix must be sampled identically.')

    if tWeightMatrix is None:
        tWeightMatrix = np.vstack((refMatrix[0,:], np.ones((refMatrix.shape[1]), dtype=np.float64)))
    elif (tWeightMatrix.shape[0] != 2) or (not np.all(refMatrix[0,:] == tWeightMatrix[0,:])):
        raise CostValueError('tWeightMatrix must be sampled at the same times as refMatrix and testMatrix.')

    if normalizeDataByDimRange:
        # dimRanges = np.transpose([np.ptp(refMatrix, axis=1) + np.ptp(testMatrix, axis=1)]) * 0.5
        dimRanges = np.transpose([np.ptp(refMatrix, axis=1)])
        np.divide(refMatrix[1:,:], dimRanges[1:], out=refMatrix[1:,:])
        np.divide(refMatrix[2:,:], dimRanges[2:], out=refMatrix[2:,:])
        np.divide(testMatrix[1:,:], dimRanges[1:], out=testMatrix[1:,:])
        np.divide(testMatrix[2:,:], dimRanges[2:], out=testMatrix[2:,:])

    refTestDifferences = np.abs(refMatrix[1:,:] - testMatrix[1:,:])
    tWeightMatrixedDistances = np.multiply(np.linalg.norm(refTestDifferences, axis=0), tWeightMatrix[1,:])
    return np.sqrt(np.sum(tWeightMatrixedDistances) / tWeightMatrixedDistances.size, dtype=np.float64)



#!python3
from __future__ import annotations
import numpy as np
from typing import List, Tuple, Any, Union, Callable
from magLabUtilities.exceptions.exceptions import SignalTypeError, SignalValueError

class SignalThread:
    def __init__(self, data:Union[np.ndarray, list]):
        # Convert various input datatypes to a Numpy array
        if isinstance(data, np.ndarray):
            self.data = data

        elif isinstance(data, list):
            self.data = np.asarray(data)

        else:
            raise SignalTypeError('Cannot create signal from supplied data type.')

        # Ensure that the Numpy array is 1-dimensional
        if len(self.data.shape) > 1:
            raise SignalTypeError('Cannot create signal from multi-dimensional array.')

    @classmethod
    def fromThreadSequence(cls, threadList):
        return cls(np.hstack([thread.data for thread in threadList]))

    @property
    def isIncreasing(self):
        return np.all(np.diff(self.data) > 0.0)

    @property
    def isMonotonicallyIncreasing(self):
        return np.all(np.diff(self.data) >= 0.0)

    @property
    def isDecreasing(self):
        return np.all(np.diff(self.data) < 0.0)

    @property
    def isMonotonicallyDecreasing(self):
        return np.all(np.diff(self.data) <= 0.0)

    @property
    def length(self):
        return self.data.size    

class Signal:
    def __init__(self, signalConstructorType:str, constructorTuple:Tuple):
        if signalConstructorType == 'like':
            # constructorTuple = [SignalThread, SignalThread]
            self.independentThread = constructorTuple[0]
            self.dependentThread = constructorTuple[1]
            self.signalType = 'discrete'

        elif signalConstructorType == 'fromThreadPair':
            # constructorTuple = [SignalThread, SignalThread]
            self.independentThread = constructorTuple[0]
            self.dependentThread = constructorTuple[1]
            self.signalType = 'discrete'

        elif signalConstructorType == 'fromSingleThread':
            # constructorTuple = [SignalThread, SignalThread]
            self.independentThread = constructorTuple[0]
            self.dependentThread = constructorTuple[1]
            self.signalType = 'discrete'

        elif signalConstructorType == 'fromFunctionGenerator':
            # constructorTuple = [SignalThread, SignalThread]
            self.independentThread = constructorTuple[0]
            self.dependentThread = constructorTuple[1]
            self.signalType = 'continuous'

        elif signalConstructorType == 'fromSignalSequence':
            # constructorTuple = [SignalThread, SignalThread]
            self.independentThread = constructorTuple[0]
            self.dependentThread = constructorTuple[1]
            self.signalType = 'discrete'

    @classmethod
    def like(cls, signal:Signal):
        return cls('like', (np.zeros_like(signal.independentThread.data), signal.dependentThread.data))

    @classmethod
    def fromThreadPair(cls, independentThread:SignalThread, dependentThread:SignalThread):
        # Check independentThread type
        if isinstance(independentThread, SignalThread):
            independentThread = independentThread
        else:
            raise SignalTypeError('Independent signal thread must be of type "SignalThread".')

        # Check dependentThread type
        if isinstance(dependentThread, SignalThread):
            dependentThread = dependentThread
        else:
            raise SignalTypeError('Dependent signal thread must be of type "SignalThread".')

        # Check that the independent and dependent signal threads are the same length
        if independentThread.length != dependentThread.length:
            raise SignalValueError('Independent and Dependent signal threads must be the same length.')
            
        # Check that the dependent signal thread is strictly increasing (avoids problems with calculus operations)
        if not dependentThread.isIncreasing:
            raise SignalValueError('Dependent signal thread must be strictly increasing.')

        # Prepare constructor call
        signalConstructorType = 'fromThreadPair'
        return cls(signalConstructorType, (independentThread, dependentThread))

    @classmethod
    def fromSingleThread(cls, independentThread:SignalThread, parameterizationMethod:str, tStart:np.float64=0.0, totalArcLength:np.float64=None):
        # Check independentThread type
        if isinstance(independentThread, SignalThread):
            independentThread = independentThread
        else:
            raise SignalTypeError('Independent signal thread must be of type "SignalThread".')

        # Parameterize the independent thread
        if parameterizationMethod == 'indices':
            dependentThread = SignalThread(np.arange(0, independentThread.length, step=1))
        elif parameterizationMethod == 'arcLength':
            dependentThread = SignalThread(Signal.arcLength1D(independentThread.data, prependArcLength=tStart, totalArcLength=totalArcLength))
        else:
            raise SignalValueError('Parameterization method ''%s'' is not recognized.')

        # Prepare constructor call
        signalConstructorType = 'fromSingleThread'
        return cls(signalConstructorType, (independentThread, dependentThread))

    @classmethod
    def fromFunctionGenerator(cls, functionGenerator, parameterizationMethod=Tuple[str, Any]):
        # # Check function generator type
        # if not isinstance(functionGenerator, SignalSequence):
        #     raise SignalTypeError('functionGenerator must be of type "FunctionGenerator".')

        # Create a discrete version of the function
        if parameterizationMethod[0] == 'fromSignalThread':
            dependentThread = parameterizationMethod[1]
            independentThread = functionGenerator.toSignalThread(dependentThread)
        else:
            raise SignalTypeError('Unrecognized parameterization method.')

        # Prepare constructor call
        signalConstructorType = 'fromFunctionGenerator'
        return cls(signalConstructorType, (independentThread, dependentThread))

    @classmethod
    def fromSignalSequence(cls, signalList:List[Signal]) -> Signal:
        independentThread = SignalThread.fromThreadSequence(signal.independentThread for signal in signalList)
        dependentThread = SignalThread.fromThreadSequence(signal.dependentThread for signal in signalList)

        # Prepare constructor call
        signalConstructorType = 'fromSignalSequence'
        return cls(signalConstructorType, (independentThread, dependentThread))

    @staticmethod
    def arcLength1D(data:np.ndarray, prependArcLength:np.float64=0.0, totalArcLength:np.float64=None) -> np.ndarray:
        arcLength = np.cumsum(np.abs(np.ediff1d(data, to_begin=0.0)))
        if totalArcLength is not None:
            np.multiply(arcLength, totalArcLength / arcLength[-1], out=arcLength)
        np.add(arcLength, prependArcLength, out=arcLength)
        return arcLength

    def sample(self, tThread:SignalThread, interpolationMethod:Callable[[Signal, SignalThread], Signal]) -> Signal:
        return interpolationMethod(self, tThread)

class SignalBundle:
    def __init__(self):
        self.signals = {}

    @classmethod
    def fromSignalBundleSequence(cls, signalBundleList:List[SignalBundle]) -> SignalBundle:
        signalDict = {}
        for signalBundle in signalBundleList:
            for signalKey in signalBundle.signals.keys():
                if signalKey in signalDict:
                    signalDict[signalKey].append(signalBundle.signals[signalKey])
                else:
                    signalDict[signalKey] = [signalBundle.signals[signalKey]]

        compiledBundle = cls()
        for signalKey in signalDict.keys():
            compiledBundle.addSignal(signalKey, Signal.fromSignalSequence(signalDict[signalKey]))

        return compiledBundle

    @classmethod
    def fromSignalBundleArray(cls, signalBundleArray:np.ndarray, signalNames:List[str]) -> SignalBundle:
        if not (len(signalNames) == signalBundleArray.shape[0]-1):
            raise SignalValueError('Number of signal names must match the number of indpendent lines in signalBundleArray.')
        newSignalBundle = cls()
        for i, signalName in enumerate(signalNames):
            newSignalBundle.addSignal(signalName, Signal.fromThreadPair(SignalThread(signalBundleArray[i+1,:]), SignalThread(signalBundleArray[0,:])))

        return newSignalBundle

    # Computes the n-dimensional arclength along a parametric path defined by one or more Signals.
    # This function should be called using the output of SignalBundle.sample() to ensure that each
    # independent signal is sampled at the same times.
    @staticmethod
    def arcLengthND(dataBundleArray:np.ndarray, prependArcLength:np.float64=0.0, totalArcLength:np.float64=None, normalizeAxes=False) -> np.ndarray:
        # Normalize and scale each independent variable to the same range (-1,1)
        if normalizeAxes:
            normalizedArray = np.empty((dataBundleArray.shape[0]-1, dataBundleArray.shape[1]))
            for signalRow in range(0, dataBundleArray.shape[0]-1):
                normalizedArray[signalRow] = np.interp(dataBundleArray[signalRow+1], (dataBundleArray[signalRow+1].min(), dataBundleArray[signalRow+1].max()), (-1,1))
        # Calculate n-dim arc lengths and cumulative lengths
        np.cumsum(np.linalg.norm(np.diff(normalizedArray, axis=1), axis=0), out=dataBundleArray[0,1:])
        dataBundleArray[0,0] = 0.0
        # Normalize arc length by totalArcLength if given
        if totalArcLength is not None:
            np.multiply(dataBundleArray[0,:], totalArcLength / dataBundleArray[0,-1], out=dataBundleArray[0,:])
        # Offset arc length by prependArcLength if given
        np.add(dataBundleArray[0,:], prependArcLength, out=dataBundleArray[0,:])
        return dataBundleArray

    def addSignal(self, name:str, signal:Signal) -> None:
        if name in self.signals.keys():
            raise SignalValueError('Cannot add Signal (%s) to bundle. "%s" already exists.')

        self.signals[name] = signal

    def sample(self, tThread:SignalThread, signalInterpList:List[Tuple[str, Callable[[Signal, SignalThread], Signal]]]) -> np.ndarray:
        sampledSignalList = []
        for signalInterp in signalInterpList:
            if len(signalInterp) != 2:
                raise SignalValueError('Must provide Signal name and interpolation method.')
            if signalInterp[0] not in self.signals.keys():
                raise SignalValueError('Signal ''%s'' does not exist in this bundle.')

            sampledSignalList.append(self.signals[signalInterp[0]].sample(tThread, signalInterp[1]))

        return np.vstack((tThread.data, np.vstack([sampledSignal.independentThread.data for sampledSignal in sampledSignalList])))

#\todo - SignalSequence tThread is jenky...
class SignalSequence:
    def __init__(self, t0:np.float64=np.float64(0.0)):
        self.functionList = []
        self.duration = t0

    def appendFunction(self, function:Callable[[SignalThread], Signal], functionT0:np.float64, functionT1:np.float64) -> None:
        if functionT0 > functionT1:
            raise SignalValueError('functionT0 must be less than or equal to functionT1')

        self.functionList.append((function, functionT0, functionT1-functionT0))
        self.duration += abs(functionT1 - functionT0)

    # def evaluate(self, tThread:SignalThread) -> Signal:
    #     t = np.float64(0.0)
    #     if not tThread.isIncreasing:
    #         raise SignalValueError('tThread must be increasing.')

    #     signalList = []
    #     functionRegion = None
    #     for function in self.functionList:
    #         if function is self.functionList[-1]:
    #             functionRegion = np.where(np.logical_and(tThread.data >= t, tThread.data <= t+function[2]))[0]
    #         else:
    #             functionRegion = np.where(np.logical_and(tThread.data >= t, tThread.data < t+function[2]))[0]
    #         t += function[2]

    #         regionTThread = SignalThread(tThread.data[functionRegion[0]:functionRegion[-1]+1] - tThread.data[functionRegion[0]] + function[1])
    #         signalList.append(function[0](regionTThread))

    #     return Signal.fromSignalSequence(signalList, tThread)



#!python3

# \name
#   hysteresis curve utilities
#
# \description
#   Module for common manipulations of hysteresis curves
#
# \notes
#
# \todo
#
# \revised
#   Mark Travers 6/16/2020      - Original construction

import numpy as np
from magLabUtilities.signalutilities.signals import SignalThread, Signal, SignalBundle
from magLabUtilities.signalutilities.calculus import integralTrapQuadrature
from magLabUtilities.exceptions.exceptions import SignalTypeError

class HysteresisSignalBundle(SignalBundle):
    def __init__(self, signalBundle:SignalBundle=None):
        super().__init__()
        if isinstance(signalBundle, SignalBundle):
            self.signals = signalBundle.signals
        elif signalBundle is not None:
            raise SignalTypeError('Cannot cast %s to HysteresisSignalBundle.' % str(type(signalBundle)))

    @property
    def mhComplete(self):
        if 'M' in self.signals.keys() and 'H' in self.signals.keys():
            return True
        else:
            return False

    @property
    def xmComplete(self):
        if 'X' in self.signals.keys() and 'M' in self.signals.keys():
            return True
        else:
            return False

    @property
    def xmhComplete(self):
        if 'X' in self.signals.keys() and 'M' in self.signals.keys() and 'H' in self.signals.keys():
            return True
        else:
            return False

    @property
    def xRevmComplete(self):
        if 'Xrev' in self.signals.keys() and 'M' in self.signals.keys():
            return True
        else:
            return False

class XExpQAGedney061720:
    def __init__(self, xInit:np.float64, hCoercive:np.float64, mSat:np.float64, hCoop:np.float64, hAnh:np.float64, xcPow:np.float64, mRev:np.float64, virginMTolerance:np.float64):
        self.xInit = np.float64(xInit)
        self.hCoercive = np.float64(hCoercive)
        self.mSat = np.float64(mSat)
        self.hCoop = np.float64(hCoop)
        self.hAnh = np.float64(hAnh)
        self.xcPow = np.float64(xcPow)
        self.mRev = np.float64(mRev)
        self.virginMTolerance = np.float64(virginMTolerance)

    def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
        if abs(self.mRev) <= self.virginMTolerance:
            absDM = np.abs(mSignal.independentThread.data)
            hCTerm = self.hCoercive
        else:
            absDM = np.abs(mSignal.independentThread.data - self.mRev)
            hCTerm = self.hCoop

        xr = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2)
        xc = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), self.xcPow)
        xRev = self.xInit * xr
        mSatMM = np.power(self.mSat, 2) - np.power(mSignal.independentThread.data, 2)

        exponent = absDM / (self.xInit * (hCTerm + (self.hAnh*self.mSat*absDM)/(xc*mSatMM)))
        xSignal = Signal.fromThreadPair(SignalThread(xRev * np.exp(exponent)), mSignal.dependentThread)

        xOfMBundle = HysteresisSignalBundle()
        xOfMBundle.addSignal('M', mSignal)
        xOfMBundle.addSignal('X', xSignal)

        return xOfMBundle

class XExpQACarl060820:
    def __init__(self, xInit:np.float64, hCoercive:np.float64, mSat:np.float64, hCoop:np.float64, hAnh:np.float64, xcPow:np.float64, mRev:np.float64, virginMTolerance:np.float64):
        self.xInit = np.float64(xInit)
        self.hCoercive = np.float64(hCoercive)
        self.mSat = np.float64(mSat)
        self.hCoop = np.float64(hCoop)
        self.hAnh = np.float64(hAnh)
        self.xcPow = np.float64(xcPow)
        self.mRev = np.float64(mRev)
        self.virginMTolerance = np.float64(virginMTolerance)

    def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
        if abs(self.mRev) <= self.virginMTolerance:
            absDM = np.abs(mSignal.independentThread.data)
            hCTerm = self.hCoercive / (2.0 * absDM)
        else:
            absDM = np.abs(mSignal.independentThread.data - self.mRev)
            hCTerm = self.hCoop / (2.0 * absDM)

        xr = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2)
        xc = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), self.xcPow)
        xRev = self.xInit * xr

        hAnhTerm = self.hAnh / (self.mSat * np.power(1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2), 2))
        exponent = xc / (self.xInit * (hCTerm + hAnhTerm))
        xSignal = Signal.fromThreadPair(SignalThread(xRev * np.exp(exponent)), mSignal.dependentThread)

        xOfMBundle = HysteresisSignalBundle()
        xOfMBundle.addSignal('M', mSignal)
        xOfMBundle.addSignal('X', xSignal)

        return xOfMBundle

class XExpGedney060820:
    def __init__(self, xInit:np.float64, hCoercive:np.float64, mSat:np.float64, hCoop:np.float64, hAnh:np.float64, xcPow:np.float64, mRev:np.float64, virginMTolerance:np.float64):
        self.xInit = np.float64(xInit)
        self.hCoercive = np.float64(hCoercive)
        self.mSat = np.float64(mSat)
        self.hCoop = np.float64(hCoop)
        self.hAnh = np.float64(hAnh)
        self.xcPow = np.float64(xcPow)
        self.mRev = np.float64(mRev)
        self.virginMTolerance = np.float64(virginMTolerance)

    def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
        if abs(self.mRev) <= self.virginMTolerance:
            hCTerm = self.hCoercive / 2.0
        else:
            hCTerm = self.hCoop

        absDM = np.abs(mSignal.independentThread.data - self.mRev)
        xr = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2)
        xc = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), self.xcPow)
        xRev = self.xInit * xr

        mSatMM = np.array(np.zeros_like(mSignal.independentThread.data, np.float64))
        gtIndices = np.where(mSignal.independentThread.data - self.mRev >= 0)[0]
        mSatMM[gtIndices] = self.mSat - mSignal.independentThread.data[gtIndices]
        ltIndices = np.where(mSignal.independentThread.data - self.mRev < 0)
        mSatMM[ltIndices] = self.mSat + mSignal.independentThread.data[ltIndices]

        exponent = absDM / (self.xInit * (hCTerm/xc + (self.hAnh*absDM)/mSatMM))
        xSignal = Signal.fromThreadPair(SignalThread(xRev * np.exp(exponent)), mSignal.dependentThread)

        xOfMBundle = HysteresisSignalBundle()
        xOfMBundle.addSignal('M', mSignal)
        xOfMBundle.addSignal('X', xSignal)

        return xOfMBundle

class XExpMremGedney070720:
    def __init__(self, xInit:float, hCoercive:float, mRem:float, mNuc:float, mSat:float, hAnh:float, mRev:float, virginMTolerance:float):
        self.xInit = np.float64(xInit)
        self.hCoercive = np.float64(hCoercive)
        self.mRem = np.float64(mRem)
        self.mNuc = np.float64(mNuc)
        self.mSat = np.float64(mSat)
        self.hAnh = np.float64(hAnh)
        self.mRev = np.float64(mRev)
        self.virginMTolerance = np.float64(virginMTolerance)

    def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
        # Detect if this is the virgin curve
        if abs(self.mRev) <= abs(self.virginMTolerance):
            absDM = np.abs(mSignal.independentThread.data)
            hCTerm = self.hCoercive / (2.0 * absDM)
        # Detect if this is a positive reversal
        elif self.mRev > abs(self.virginMTolerance):
            absDM = np.abs(mSignal.independentThread.data - self.mRev)

            aboveNucIndices = np.where(mSignal.independentThread.data < -abs(self.mNuc))
            betweenIndices = np.where(np.logical_and(mSignal.independentThread.data > -abs(self.mNuc), mSignal.independentThread.data < -abs(self.mRem)))
            belowRemIndices = np.where(mSignal.independentThread.data > -abs(self.mRem))
            hCoop = np.array(np.empty_like(mSignal.independentThread.data))
            hCoop[aboveNucIndices] = self.hCoercive
            hCoop[betweenIndices] = self.hCoercive * np.abs(mSignal.independentThread.data[betweenIndices] - abs(self.mRem)) / (abs(self.mNuc) - abs(self.mRem))
            hCoop[belowRemIndices] = 0.0

            hCTerm = hCoop / (4.0 * absDM)
        # Detect if this is a negative reversal
        elif self.mRev < -abs(self.virginMTolerance):
            absDM = np.abs(mSignal.independentThread.data - self.mRev)

            aboveNucIndices = np.where(mSignal.independentThread.data > abs(self.mNuc))
            betweenIndices = np.where(np.logical_and(mSignal.independentThread.data < abs(self.mNuc), mSignal.independentThread.data > abs(self.mRem)))
            belowRemIndices = np.where(mSignal.independentThread.data < abs(self.mRem))
            hCoop = np.array(np.empty_like(mSignal.independentThread.data))
            hCoop[aboveNucIndices] = self.hCoercive
            hCoop[betweenIndices] = self.hCoercive * np.abs(mSignal.independentThread.data[betweenIndices] + abs(self.mRem)) / (abs(self.mNuc) - abs(self.mRem))
            hCoop[belowRemIndices] = 0.0

            hCTerm = hCoop / (4.0 * absDM)
        
        xr = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2)
        xRev = self.xInit * xr

        hAnhTerm = self.hAnh / (self.mSat * np.power(1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2), 2))

        exponent = 1.0 / (self.xInit * (hCTerm + hAnhTerm))
        xSignal = Signal.fromThreadPair(SignalThread(xRev * np.exp(exponent)), mSignal.dependentThread)

        xOfMBundle = HysteresisSignalBundle()
        xOfMBundle.addSignal('M', mSignal)
        xOfMBundle.addSignal('X', xSignal)

        return xOfMBundle

# class XExpTravers071520:
#     def __init__(self, xInit, hRev, mRev, hCoercive, mCoercive, hRem, mRem, hAnh, mSat, excitationDirection):
#         self.xInit = np.float64(xInit)
#         self.hRev = np.float64(hRev)
#         self.mRev = np.float64(mRev)
#         self.hCoercive = np.float64(hCoercive)
#         self.mCoercive = np.float64(mCoercive)
#         self.hRem = np.float64(hRem)
#         self.mRem = np.float64(mRem)
#         self.hAnh = np.float64(hAnh)
#         self.mSat = np.float64(mSat)

#         self.excitationDirection = excitationDirection

#     def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
#         xRev = self.xInit * (1.0 * np.square(mSignal.independentThread.data / self.mSat))

#         if self.excitationDirection == 'increasing':
#             revTerm = self.hRev / np.abs(mSignal.independentThread.data - self.mRev)
#             coerciveTerm = self.hCoercive / np.abs(mSignal.independentThread.data - self.mCoercive)
#             remTerm = self.hRem / np.abs(mSignal.independentThread.data - self.mRem)
#             anhTerm = self.hAnh / (self.mSat - np.abs())

# Does not function correctly
class XExpGedney071720:
    def __init__(self, xInit:float, hCoercive:float, hNuc:float, mNuc:float, mSat:float, hCoop:float, hAnh:float, hRev:float, mRev:float, curveRegion:str):
        self.xInit = np.float64(abs(xInit))
        self.hCoercive = np.float64(abs(hCoercive))
        self.hNuc = np.float64(abs(hNuc))
        self.mNuc = np.float64(abs(mNuc))
        self.mSat = np.float64(abs(mSat))
        self.hCoop = np.float64(abs(hCoop))
        self.hAnh = np.float64(abs(hAnh))

        self.hRev = np.float64(hRev)
        self.mRev = np.float64(mRev)
        self.curveRegion = curveRegion

    def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
        xSignal = Signal.fromThreadPair(SignalThread(np.zeros_like(mSignal.independentThread.data)), mSignal.dependentThread)
        hSignal = Signal.fromThreadPair(SignalThread(np.zeros_like(mSignal.independentThread.data)), mSignal.dependentThread)

        mSignal.independentThread.data[0] = self.mRev

        xr = 1.0 - np.power(mSignal.independentThread.data / self.mSat, 2)
        xRev = self.xInit * xr
        absDM = np.abs(mSignal.independentThread.data - self.mRev)
        hAnhTerm = (self.hAnh * absDM) / (self.mSat * np.power(1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2), 2))

        for i in range(xSignal.independentThread.length):
            if abs(mSignal.independentThread.data[i]) > self.mNuc:
                xSignal.independentThread.data[i] = xRev[i]
            else:
                if self.curveRegion == 'virgin':
                    hCTerm = self.hCoercive * 0.5
                elif self.curveRegion == 'reversal':
                    if self.hRev < -self.hNuc:
                        self.hRev = -self.hNuc
                    elif self.hRev > self.hNuc:
                        self.hRev = self.hNuc

                    x = self.hRev - mSignal.independentThread.data[i]

                    if abs(x) > abs(self.hRev):
                        x = 1.0
                    else:
                        x = abs(x) / abs(self.hRev)
                    hCTerm = (self.hCoop - self.hCoercive) * (x**2) + self.hCoercive

                exponent = absDM[i] / (self.xInit * (hCTerm + hAnhTerm))
                xSignal.independentThread.data[i] = xRev[i] * np.exp(exponent)[0]

            if i+1 < hSignal.independentThread.length:
                hSignal.independentThread.data[i+1] = hSignal.independentThread.data[i] + (mSignal.independentThread.data[i+1] - mSignal.independentThread.data[i]) / xSignal.independentThread.data[i]

        hysteresisBundle = HysteresisSignalBundle()
        hysteresisBundle.addSignal('M', mSignal)
        hysteresisBundle.addSignal('H', hSignal)
        hysteresisBundle.addSignal('X', xSignal)

        xRevBundle = HysteresisSignalBundle()
        xRevBundle.addSignal('M', mSignal)
        xRevBundle.addSignal('X', Signal.fromThreadPair(SignalThread(xRev), mSignal.dependentThread))

        return hysteresisBundle, xRevBundle

# Does not function correctly
class XExpGedneyAlt071620:
    def __init__(self, xInit:float, hCoercive:float, hCoop:float, mNuc:float, mSat:float, hAnh:float, mRev:float, curveRegion:str):
        self.xInit = np.float64(abs(xInit))
        self.hCoercive = np.float64(abs(hCoercive))
        self.hCoop = np.float(abs(hCoop))
        self.mNuc = np.float64(abs(mNuc))
        self.mSat = np.float64(abs(mSat))
        self.hAnh = np.float64(abs(hAnh))
        self.mRev = np.float64(mRev)
        self.curveRegion = curveRegion

    def evaluate(self, mSignal:Signal) -> HysteresisSignalBundle:
        if self.curveRegion == 'virgin':
            absDM = np.abs(mSignal.independentThread.data)
            hCTerm = self.hCoercive / 2.0
        else:
            absDM = np.abs(mSignal.independentThread.data - self.mRev)
            
            if self.mRev < -self.mNuc:
                self.mRev = -self.mNuc
            elif self.mRev > self.mNuc:
                self.mRev = self.mNuc

            belowRemIndices = np.where(np.abs(self.mRev - mSignal.independentThread.data) < abs(self.mRev))
            aboveRemIndices = np.where(np.abs(self.mRev - mSignal.independentThread.data) >= abs(self.mRev))

            hCTerm = np.array(np.empty_like(mSignal.independentThread.data))
            hCTerm[belowRemIndices] = (self.hCoop - self.hCoercive) * ((self.mRev - mSignal.independentThread.data[belowRemIndices]) / self.mNuc) + self.hCoercive
            hCTerm[aboveRemIndices] = (self.hCoop - self.hCoercive) * (self.mRev / self.mNuc) + self.hCoercive
        
        xr = 1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2)
        xRev = self.xInit * xr

        hAnhTerm = (self.hAnh * absDM) / (self.mSat * np.power(1.0 - np.power(np.abs(mSignal.independentThread.data / self.mSat), 2), 2))

        exponent = absDM / (self.xInit * (hCTerm + hAnhTerm))
        xSignal = Signal.fromThreadPair(SignalThread(xRev * np.exp(exponent)), mSignal.dependentThread)

        xOfMBundle = HysteresisSignalBundle()
        xOfMBundle.addSignal('M', mSignal)
        xOfMBundle.addSignal('X', xSignal)

        return xOfMBundle

class XExpOfHGedney071720:
    def __init__(self, xInit:float, hCoercive:float, hNuc:float, mNuc:float, mSat:float, hCoop:float, hAnh:float):
        self.xInit = np.float64(abs(xInit))
        self.hCoercive = np.float64(abs(hCoercive))
        self.hNuc = np.float64(abs(hNuc))
        self.mNuc = np.float64(abs(mNuc))
        self.mSat = np.float64(abs(mSat))
        self.hCoop = np.float64(abs(hCoop))
        self.hAnh = np.float64(abs(hAnh))

    def evaluate(self, hSignal:Signal, hRev:float, mRev:float, curveRegion:str) -> HysteresisSignalBundle:
        

        hRev = np.float64(hRev)
        mRev = np.float64(mRev)
        curveRegion = curveRegion

        xSignal = Signal.fromThreadPair(SignalThread(np.zeros_like(hSignal.independentThread.data)), hSignal.dependentThread)
        mSignal = Signal.fromThreadPair(SignalThread(np.zeros_like(hSignal.independentThread.data)), hSignal.dependentThread)
        xRevSignal = Signal.fromThreadPair(SignalThread(np.zeros_like(hSignal.independentThread.data)), hSignal.dependentThread)

        mSignal.independentThread.data[0] = mRev

        if mRev > self.mNuc:
            mRev = self.mNuc
        elif mRev <= -self.mNuc:
            mRev = -self.mNuc

        if hRev < -self.hNuc:
            hRev = -self.hNuc
        elif hRev >= self.hNuc:
            hRev = self.hNuc

        for i in range(xSignal.independentThread.length):
            xr = 1.0 - np.power(mSignal.independentThread.data[i] / self.mSat, 2)
            xRevSignal.independentThread.data[i] = self.xInit * xr
            absDM = np.abs(mSignal.independentThread.data[i] - mRev)
            hAnhTerm = (self.hAnh * absDM) / (self.mSat * np.power(1.0 - np.power(np.abs(mSignal.independentThread.data[i] / self.mSat), 2), 2))

            if abs(mSignal.independentThread.data[i]) > self.mNuc:
                xSignal.independentThread.data[i] = xRevSignal.independentThread.data[i]
            else:
                if curveRegion == 'virgin':
                    hCTerm = self.hCoercive * 0.5
                elif curveRegion == 'reversal':
                    x = hRev - hSignal.independentThread.data[i]

                    if abs(x) > abs(hRev):
                        x = 1.0
                    else:
                        x = abs(x) / abs(hRev)
                    hCTerm = (self.hCoop - self.hCoercive) * (x**1) + self.hCoercive

                exponent = absDM / (self.xInit * (hCTerm + hAnhTerm))
                xSignal.independentThread.data[i] = xRevSignal.independentThread.data[i] * np.exp(exponent)

            if i+1 < hSignal.independentThread.length:
                mSignal.independentThread.data[i+1] = mSignal.independentThread.data[i] + (hSignal.independentThread.data[i+1] - hSignal.independentThread.data[i]) * xSignal.independentThread.data[i]

        hysteresisBundle = HysteresisSignalBundle()
        hysteresisBundle.addSignal('M', mSignal)
        hysteresisBundle.addSignal('H', hSignal)
        hysteresisBundle.addSignal('X', xSignal)
        hysteresisBundle.addSignal('Xrev', xRevSignal)

        return hysteresisBundle

class XExpGendey101620:
    def __init__(self, xInit:np.float64, mSat:np.float64, mNuc:np.float64, hCoercive:np.float64, hAnh:np.float64, hCoop:np.float64):
        self.xInit = xInit
        self.mSat = mSat
        self.mNuc = mNuc
        self.hCoercive = hCoercive
        self.hAnh = hAnh
        self.hCoop = hCoop

    def evaluate(self, mSignal:Signal, mRev:np.float64, hRev:np.float64, curveRegion:str):
        # Calculate intermediate expressions
        m = np.divide(mSignal.independentThread.data, self.mSat)
        xR = 1.0 - np.power(m, 2)
        dM = np.abs(np.subtract(mSignal.independentThread.data, mRev))

        # Setup constants which change for virgin and reversal regions in and out of nucleation
        if curveRegion == 'virgin':
            hCP = self.hCoercive * 0.5
        elif curveRegion == 'reversal':
            if abs(mRev) > self.mNuc:
                if mRev < 0:
                    mRev = -self.mNuc
                else:
                    mRev = self.mNuc
            hCP = self.hCoop

        # Set up Xrm and Xrc
        xRM = np.empty_like(mSignal.independentThread.data)
        xRC = np.empty_like(mSignal.independentThread.data)
        mPosIndices = np.asarray(mSignal.independentThread.data > 0.0).nonzero()
        mNegIndices = np.asarray(mSignal.independentThread.data <= 0.0).nonzero()
        if mRev > 0.0:
            xRM[mPosIndices] = 1.0 + np.power(m[mPosIndices], 2)
            xRM[mNegIndices] = xR[mNegIndices]
            xRC[mPosIndices] = xR[mPosIndices]
            xRC[mNegIndices] = 1.0 + np.power(m[mNegIndices], 2)
        else:
            xRM[mNegIndices] = 1.0 + np.power(m[mNegIndices], 2)
            xRM[mPosIndices] = xR[mPosIndices]
            xRC[mNegIndices] = xR[mNegIndices]
            xRC[mPosIndices] = 1.0 + np.power(m[mPosIndices], 2)

        # Calculate Xexp
        xExp = xR * self.xInit
        belowNucIndices = np.asarray(np.abs(mSignal.independentThread.data) <= self.mNuc).nonzero()
        exponents = dM[belowNucIndices] / (self.xInit * (np.divide(hCP,np.float_power(xRC[belowNucIndices], 2.5)) + np.divide(self.hAnh*dM[belowNucIndices], self.mSat*np.float_power(xRM[belowNucIndices], 2.0))))
        xExp[belowNucIndices] = np.multiply(xExp[belowNucIndices], np.exp(exponents))

        # Calculate H
        # hThread = integralTrapQuadrature(1.0/xExp, mSignal.independentThread.data, hRev)
        hThread = integralTrapQuadrature(1.0/xExp, mSignal.independentThread.data, hRev)

        # Compile SignalBundle
        signalBundleArray = np.vstack((mSignal.dependentThread.data, mSignal.independentThread.data, hThread.data, xExp))
        return HysteresisSignalBundle(SignalBundle.fromSignalBundleArray(signalBundleArray, ['M','H','X']))



#!python3

import numpy as np
import pandas as pd
import re
from typing import List
from magLabUtilities.signalutilities.signals import SignalThread, Signal, SignalBundle
from magLabUtilities.exceptions.exceptions import FileIOValueError

def importFromXlsx(fp:str, sheetName:str, headerRow:int, excelDataColumns:str, excelTColumn:str=None, dataColumnNames:List[str]=None) -> SignalBundle:
        dataColumns = pd.read_excel(fp, sheet_name=sheetName, header=headerRow-1, usecols=excelDataColumns, dtype=np.float64)

        if excelTColumn is not None:
            if not excelTColumn.isalpha():
                raise FileIOValueError('Only one tColumn supported.')
            tColumn = pd.read_excel(fp, sheet_name=sheetName, header=None, skiprows=headerRow, usecols=excelTColumn, dtype=np.float64)

        if dataColumnNames is not None:
            if len(dataColumnNames) != len(dataColumns.columns):
                raise FileIOValueError('Must specify the same number of column names as number of columns requested.')

        fileSignalBundle = SignalBundle()

        for i, dataColumnKey in enumerate(dataColumns.columns):
            dataColumn = dataColumns[dataColumnKey].dropna()
            independentThread = SignalThread(dataColumn.values)
            if excelTColumn is None:
                dependentThread = SignalThread(dataColumn.index.values)
            else:
                dependentThread = SignalThread(tColumn.values[dataColumn.index.values].T[0])

            if dataColumnNames is None:
                fileSignalBundle.addSignal(dataColumnKey, Signal.fromThreadPair(independentThread, dependentThread))
            else:
                fileSignalBundle.addSignal(dataColumnNames[i], Signal.fromThreadPair(independentThread, dependentThread))

        return fileSignalBundle

def exportSignalArrayToXlsx(fp:str, exportBundleArray:np.ndarray, columnNames:List[str], sheetName:str):
    arrayDataFrame = pd.DataFrame(exportBundleArray.T)
    arrayDataFrame.to_excel(fp, sheet_name=sheetName, header=columnNames, index=False)



#!python3

from magstrommanager.testCase import MagstromTestCase
from magstrommanager.files import MagstromFiles

from optimizers import GradientDescent
from optimizers.utilities import TestManager, ParameterSpace, LossFunction

from dsp.utilities import DspChain, DataFrame
from dsp.calculus import integral, derivative
from dsp.interpolation import legendreInterpolation
from dsp.conditioners import parameterize, weightRegions, scale

from datafileutilities import decoders, encoders






#!python3

import numpy as np
from magLabUtilities.signalutilities.signals import SignalThread, Signal
from magLabUtilities.signalutilities.calculus import integralTrapQuadrature
# from internal.terminalInterface import printMsg
# from internal.exceptions import IncompatibleArgumentError, DataShapeError

def nearestPoint(interpSignal:Signal, tThread:SignalThread) -> Signal:
    tDistanceMatrix = np.vstack([interpSignal.dependentThread.data]*tThread.length)
    np.abs(tDistanceMatrix - np.transpose([tThread.data]), out=tDistanceMatrix)
    interpIndices = np.argmin(tDistanceMatrix, axis=1)

    independentThread = SignalThread(interpSignal.independentThread.data[interpIndices])
    dependentThread = SignalThread(interpSignal.dependentThread.data[interpIndices])

    return Signal.fromThreadPair(independentThread, dependentThread)

class Legendre:
    def __init__(self, interpRadius:int, legendreOrder:int):
        self.interpRadius = interpRadius
        self.legendreOrder = legendreOrder

    # Ouputs the value of the first 'order' orders of Legendre Polynomials evaluated at x
    @staticmethod
    def legendrePolynomial(x:float, order:int) -> np.ndarray:
        # Input validation
        if(order < 0):
            pass
            # raise IncompatibleArgumentError('Legendre polynomial cannot be order %d.' % order) \todo - make exception
        # Initialize first two legendre coefficients
        p = np.zeros((order + 1))
        p[0] = 1
        if(order < 1):
            return p
        p[1] = x

        # Calculate remaining legendre coefficients
        for i in range(2,order+1):
            p[i] = ((2.0 * i - 1.0)*x*p[i-1] - (i - 1.0)*p[i-2]) / i
        return p

    def interpolate(self, interpSignal:Signal, tThread:SignalThread) -> Signal:
        tDistanceMatrix = np.vstack([interpSignal.dependentThread.data]*tThread.length)
        np.abs(tDistanceMatrix - np.transpose([tThread.data]), out=tDistanceMatrix)
        interpIndices = np.argmin(tDistanceMatrix, axis=1)

        outputData = np.array(np.zeros_like(tThread.data))

        for i, interpIndex in enumerate(interpIndices):
            if interpIndex - self.interpRadius < 0:
                i0 = 0
                i1 = 2 * self.interpRadius + 1
            elif interpIndex + self.interpRadius > interpSignal.dependentThread.data.size:
                i0 = interpSignal.dependentThread.data.size - 1 - 2*self.interpRadius
                i1 = interpSignal.dependentThread.data.size - 1
            else:
                i0 = interpIndex - self.interpRadius
                i1 = interpIndex + self.interpRadius

            # Find Legendre polynomial values at t-points from the local range of interpSignal
            interpT = interpSignal.dependentThread.data[i0:i1]
            psiTArray = (interpT - np.amin(interpT)) * 2.0 / np.ptp(interpT) - 1.0
            legValueArray = np.zeros((self.legendreOrder + 1, psiTArray.size))
            for t, psiT in enumerate(psiTArray):
                legValueArray[:,t] = Legendre.legendrePolynomial(psiT, self.legendreOrder)

            # Calculate the weighting coefficients, alpha
            interpX = interpSignal.independentThread.data[i0:i1]
            alphaArray = np.zeros((self.legendreOrder + 1, 1))
            for m in range(self.legendreOrder+1):
                # innerProduct = integralTrapQuadrature(SignalThread(legValueArray[m,:] * interpX), SignalThread(psiTArray)).independentThread.data[-1]
                innerProduct = integralTrapQuadrature(SignalThread(legValueArray[m,:] * interpX), SignalThread(interpT),c=0.0).data[-1]

                alphaArray[m] = (2.0 * m + 1.0) / (interpT[-1] - interpT[0]) * innerProduct

            # Find Legendre polynomial values at current t in tThread
            psiT = (tThread.data[i] - np.amin(interpT)) * 2.0 / np.ptp(interpT) - 1.0
            legValueArray = np.empty((self.legendreOrder + 1, 1))
            legValueArray[:,0] = Legendre.legendrePolynomial(psiT, self.legendreOrder)

            # Calculate f(t) at current t in tThread
            outputData[i] = np.sum(legValueArray * alphaArray)

        return Signal.fromThreadPair(SignalThread(outputData), tThread)

def legendreFilter(data:np.ndarray, integrationWindowSize:int, stepSize:int, legendreOrder:int, terminalMode='quiet') -> np.ndarray:
    # Setup and input validation
    data = np.array(data)
    if len(data.shape) > 1:
        pass
        # raise DataShapeError('Input array must be one-dimensional.') \todo - make an exception
    inputLength = data.shape[0]
    outputLength = int(np.ceil(inputLength / stepSize))
    outputData = np.zeros((outputLength), dtype=data.dtype)

    windowRadius = int(np.floor(integrationWindowSize / 2))
    psiTArray = np.linspace(-1.0, 1.0, num=2*windowRadius + 1)
    legValueArray = np.empty((legendreOrder + 1, psiTArray.size))
    for t in range(psiTArray.size):
        legValueArray[:,t] = Legendre.legendrePolynomial(psiTArray[t], legendreOrder)

    # Interpolate data
    for outputIndex, inputIndex in enumerate(range(0, inputLength, stepSize)):
        # printMsg('Interpolating output point: %d/%d' % (outputIndex, outputLength), terminalMode)

        # Set integration window limits and enforce the window size at each end of the data
        if inputIndex < windowRadius:
            t0 = 0
            t1 = 2 * windowRadius
        elif inputIndex >= inputLength - windowRadius:
            t0 = inputLength - 2 * windowRadius - 1
            t1 = inputLength - 1
        else:
            t0 = inputIndex - windowRadius
            t1 = inputIndex + windowRadius

        sumTerms = legValueArray * data[t0:t1+1]
        sumTerms[:,0] = sumTerms[:,0] * 0.5
        sumTerms[:,-1] = sumTerms[:,-1] * 0.5
        innerProduct = np.sum(sumTerms, axis=1)

        # alpha[:,outputIndex] = innerProduct * ((2*np.arange(legendreOrder+1) + 1) / (t1 - t0))

        psiT = (2 * inputIndex - (t0 + t1)) / (t1 - t0)

        outputData[outputIndex] = np.sum(Legendre.legendrePolynomial(psiT, legendreOrder) * innerProduct * ((2*np.arange(legendreOrder+1) + 1) / (t1 - t0)))

    return outputData





import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='magLabUtilities',
    version='1.0.4',
    author='Mark Travers',
    author_email='mark.2.travers@ucdenver.edu',
    description='Commonly used utilities in the UC Denver magnetics lab.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/MarkTravers/magLabUtilities',
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=[
        'astroid>=2.4.2'
        'dill>=0.3.3'
        'lazy-object-proxy>=1.4.3'
        'multiprocess>=0.70.11.1'
        'numpy>=1.18.3'
        'pandas>=1.2.0'
        'pathos>=0.2.7'
        'pox>=0.2.9'
        'ppft>=1.6.6.3'
        'python-dateutil>=2.8.1'
        'pytz>=2020.5'
        'six>=1.15.0'
        'typed-ast>=1.4.2'
        'typing>=3.7.4.3'
        'wrapt>=1.12.1'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6'
)